//*****************************************************************************
//    Description.....WS-Eventing for WCF
//                                
//    Author..........Roman Kiss, rkiss@pathcom.com
//    Copyright  2005 ATZ Consulting Inc. (see included license.rtf file)        
//                        
//    Date Created:    06/06/06
//
//    Date        Modified By     Description
//-----------------------------------------------------------------------------
//    06/06/06    Roman Kiss     Initial Revision
//    07/07/06    Roman Kiss     Version JuneCTP 2006
//*****************************************************************************
//  
#region References
using System;
using System.Diagnostics;
using System.Timers;
using System.Threading;
using System.Collections;
using System.Text;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.ServiceModel;
using System.Runtime.Serialization;
using System.Xml;
using System.Xml.Serialization;
#endregion

namespace RKiss.WSEventing.Adapters
{
    #region PersistedObject
    public class PersistedObject : System.Timers.Timer
    {
        ResourceDescriptor _rd;
        object _resource;
        public ResourceDescriptor ResourceDescriptor
        {
            get { return _rd; }
            set { _rd = value; }
        }
        public object Resource
        {
            get { return _resource; }
            set { _resource = value; }
        }
        public PersistedObject() { }
        public PersistedObject(ResourceDescriptor rd, object resource) { _rd = rd; _resource = resource; }
    }
    #endregion

    #region class Storage
    public class Storage 
    {
        public ReaderWriterLock rwl = new ReaderWriterLock();
        string _name;
        Hashtable _key = new Hashtable();
        Hashtable _val = new Hashtable();
        public string Name { get { return _name; } }
        public Hashtable Key { get { return _key; } }
        public Hashtable Val { get { return _val; } }
        public Storage(string name)
        {
            _name = name;
        }
        public void Purge()
        {
            _key.Clear();
            _val.Clear();
        }
    }
    #endregion

    public class MemoryStorageAdapter : IStorageAdapter
    {
        #region Events
        public event EventHandler<ResourceDeletedEventArgs> ResourceDeleted;
        #endregion

        #region Properties
        Storage _storage = null;
        string _strStorageName = string.Empty;
        object SyncRoot = new object();
        HybridDictionary _properties;
        int _maxNumberOfSubscriptions = 50;

        public HybridDictionary Properties
        {
            get { return _properties; }
        }
        public Storage Storage
        {
            get { return _storage; }
        }
        #endregion

        #region Constructors
        public MemoryStorageAdapter() :this(null)
        {
        }
        public MemoryStorageAdapter(HybridDictionary config)
        {
            _properties = config;
            Init();
        }
        #endregion

        #region IStorageAdapter Members
        public ResourceDescriptor Create<T>(ResourceDescriptor rd, T resource) where T : class
        {
            try
            {
                // resource identifiers (id, topic/key)
                string keyword = CreateStorageKeyword(rd, Properties);

                // serialize a resource to the XmlElement for any type T except an object 
                object xresource = (typeof(T) == typeof(object)) ? resource : Convert.ObjectToXml(resource) as object;
                PersistedObject persistedObject = new PersistedObject(rd, xresource);
                
                // lock the storage for writer
                _storage.rwl.AcquireWriterLock(TimeSpan.FromSeconds(10));

                if (_storage.Key.Count >= _maxNumberOfSubscriptions)
                {
                    throw new FaultException("The limit of the subscriptions expired.");
                }

                if (keyword != null)
                {
                    if (_storage.Key.Contains(keyword))
                    {
                        throw new FaultException(string.Format("The topic-key '{0}' already exist in the storage.", keyword),
                            FaultCode.CreateSenderFaultCode(WSEventing.InvalidMessage.SubCode, WSEventing.NamespaceUri));
                    }
                    _storage.Key[keyword] = rd.Id;
                }
                _storage.Val[rd.Id] = persistedObject;

                // subscription watchdog
                DateTime ldt = persistedObject.ResourceDescriptor.Expires.ToLocalTime();
                if(ldt != DateTime.MinValue || ldt != DateTime.MaxValue) 
                {
                    persistedObject.Interval = (ldt - DateTime.Now).TotalMilliseconds;
                    persistedObject.Elapsed += new ElapsedEventHandler(OnSubscriptionExpired);
                    persistedObject.AutoReset = false;
                    persistedObject.Enabled = true;
                }
                //Util.ShowMe(persistedObject.ResourceDescriptor);
                return rd;
            }
            finally
            {
                _storage.rwl.ReleaseWriterLock();
            }
        }

        public T Get<T>(ResourceDescriptor rd) where T: class
        {
            try
            {
                _storage.rwl.AcquireWriterLock(TimeSpan.FromSeconds(10));
                PersistedObject po = GetPersistedObject(rd);
                return (T)Convert.XmlToObject<T>(po.Resource);
            }
            finally
            {
                _storage.rwl.ReleaseWriterLock();
            }
        }
        public IList<T> Get<T>(string topic) where T : class
        {
            if(topic == null || topic.Length == 0)
                throw new FaultException("The topic must be defined.");

            List<T> list = new List<T>();
            try
            {
                _storage.rwl.AcquireWriterLock(TimeSpan.FromSeconds(10));  
                topic = topic.EndsWith("*") ? topic.TrimEnd('*') : string.Concat(topic, "&&");
                foreach (string key in _storage.Key.Keys)
                {
                    if (topic == null || key.StartsWith(topic))
                    {
                        PersistedObject po = _storage.Val[(UniqueId)_storage.Key[key]] as PersistedObject;
                        list.Add(Convert.XmlToObject<T>(po.Resource));
                    }
                }
                return list;
            }
            finally
            {
                _storage.rwl.ReleaseWriterLock();
            }
        }

        public T Delete<T>(UniqueId id)
        {
            try
            {
                _storage.rwl.AcquireWriterLock(TimeSpan.FromSeconds(10));
                if (_storage.Val.Contains(id) == false)
                {
                    throw new FaultException(string.Format("The identifier '{0}' doesn't exist in the storage.", id),
                        FaultCode.CreateSenderFaultCode(WSEventing.InvalidMessage.SubCode, WSEventing.NamespaceUri));
                }
                PersistedObject po = _storage.Val[id] as PersistedObject;
                string keyword = CreateStorageKeyword(po.ResourceDescriptor, null);
                _storage.Key.Remove(keyword);
                _storage.Val.Remove(id);
                return (T)Convert.XmlToObject<T>(po.Resource);
            }
            finally
            {
                _storage.rwl.ReleaseWriterLock();
            }
        }
        public T Delete<T>(ResourceDescriptor rd) where T : class
        {
            try
            {
                _storage.rwl.AcquireWriterLock(TimeSpan.FromSeconds(10));
                if (rd.Id != null)
                {
                    return (T)Delete<T>(rd.Id);
                }     

                string keyword = CreateStorageKeyword(rd, Properties);
                if (keyword != null)
                {
                    if (_storage.Key.Contains(keyword) == false)
                    {
                        throw new FaultException(string.Format("The topic-key '{0}' doesn't exist in the storage.", keyword),
                            FaultCode.CreateSenderFaultCode(WSEventing.InvalidMessage.SubCode, WSEventing.NamespaceUri));
                    }
                    UniqueId id = (UniqueId)_storage.Key[keyword];
                    PersistedObject po = _storage.Val[id] as PersistedObject;
                    _storage.Key.Remove(keyword);
                    _storage.Val.Remove(id);
                    return (T)Convert.XmlToObject<T>(po.Resource);
                }
                else
                {
                    throw new FaultException(string.Format("Wrong ResourceDescriptor", keyword),
                        FaultCode.CreateSenderFaultCode(WSEventing.InvalidMessage.SubCode, WSEventing.NamespaceUri));
                }
            }
            finally
            {
                _storage.rwl.ReleaseWriterLock();
            } 
        }
        public IList<T> Delete<T>(string topic) where T : class
        {
            if (topic == null || topic.Length == 0)
                throw new FaultException("The topic must be defined.");

            List<T> list = new List<T>();
            List<string> keys = new List<string>();
            try
            {
                _storage.rwl.AcquireWriterLock(TimeSpan.FromSeconds(10));
                topic = topic.EndsWith("*") ? topic.TrimEnd('*') : string.Concat(topic, "&&");

                // select resources by topic
                foreach (string key in _storage.Key.Keys)
                {
                    if (topic == null || key.StartsWith(topic))
                    {
                        UniqueId id = (UniqueId)_storage.Key[key];
                        PersistedObject po = _storage.Val[id] as PersistedObject;
                        keys.Add(key);
                        list.Add(Convert.XmlToObject<T>(po.Resource));
                       
                    }
                }

                // remove resources from the storage
                foreach (string k in keys)
                {
                    UniqueId id = (UniqueId)_storage.Key[k];
                    _storage.Key.Remove(k);
                    _storage.Val.Remove(id);
                }
                return list;
            }
            finally
            {
                _storage.rwl.ReleaseWriterLock();
            }
        }

        public void Put<T>(ResourceDescriptor rd, T resource) where T : class
        {
            throw new NotImplementedException("Put");
        }

        public void CreateOrUpdate<T>(ResourceDescriptor rd, T resource) where T : class 
        {
            throw new NotImplementedException("CreateOrUpdate");
        }

        public ResourceDescriptor[] GetResourceDescriptors(ResourceDescriptor rd)
        {
            throw new NotImplementedException("GetResourceDescriptors");
        }

        public void Expires(ResourceDescriptor rd)
        {
            try
            {
                _storage.rwl.AcquireWriterLock(TimeSpan.FromSeconds(10));
                PersistedObject po = GetPersistedObject(rd);
                po.Stop();
                po.ResourceDescriptor.Expires = rd.Expires;

                // subscription watchdog
                DateTime ldt = po.ResourceDescriptor.Expires.ToLocalTime();
                if (ldt != DateTime.MinValue || ldt != DateTime.MaxValue)
                {
                    po.Interval = (ldt - DateTime.Now).TotalMilliseconds;
                    po.Enabled = true;
                }         
            }
            finally
            {
                _storage.rwl.ReleaseWriterLock();
            }
        }

        public ResourceDescriptor GetResourceDescriptor(ResourceDescriptor rd)
        {
            try
            {
                _storage.rwl.AcquireWriterLock(TimeSpan.FromSeconds(10));
                PersistedObject po = GetPersistedObject(rd);
                return new ResourceDescriptor(po.ResourceDescriptor);
            }
            finally
            {
                _storage.rwl.ReleaseWriterLock();
            }
        }
        #endregion

        #region Internals
        private void Init()
        {
            // max number of subscriptions
            int.TryParse(Properties["maxsubscriptions"] as string, out _maxNumberOfSubscriptions);

            // storage name 
            _strStorageName = (string)Properties["name"];
            if (_strStorageName == null || _strStorageName.Length == 0)
            {
                throw new FaultException("Missing name of the storage");
            }

            // my domain
            AppDomain currentDomain = AppDomain.CurrentDomain;

            _storage = currentDomain.GetData(_strStorageName) as Storage;
            if (_storage == null)
            {
                lock (currentDomain.FriendlyName)
                {
                    _storage = currentDomain.GetData(_strStorageName) as Storage;
                    if (_storage == null)
                    {
                        _storage = new Storage(_strStorageName);
                        currentDomain.SetData(_strStorageName, _storage);
                    }
                }
            }
        }
        private string CreateStorageKeyword(ResourceDescriptor rd, HybridDictionary properties)
        {
            // from config
            string topic = properties == null ? null : (string)Properties["topic"];
            string key = properties == null ? null : (string)Properties["key"];

            // update ResourceDescriptor
            if (rd.Topic == null)
                rd.Topic = topic;
            if (rd.Key == null)
                rd.Key = key;
            if (rd.Key == null)
                rd.Key = rd.Id.ToString();

            string keyword = null;
            if (rd.Key != null && rd.Topic != null)
                keyword = string.Format("{0}&&{1}", rd.Topic, rd.Key);
            else if (rd.Topic != null && rd.Key == null)
                keyword = rd.Topic;

            return keyword;
        }
        private PersistedObject GetPersistedObject(ResourceDescriptor rd)
        {
            PersistedObject po = null;
            string keyword = CreateStorageKeyword(rd, Properties);
            if (rd.Id != null)
            {
                if (_storage.Val.Contains(rd.Id))
                {
                    po = _storage.Val[rd.Id] as PersistedObject;
                }
                else
                {
                    throw new FaultException(string.Format("The identifier '{0}' doesn't exist in the storage.", rd.Id),
                        FaultCode.CreateSenderFaultCode(WSEventing.InvalidMessage.SubCode, WSEventing.NamespaceUri));
                }
            }
            else if (keyword != null)
            {
                if (_storage.Key.Contains(keyword) == false)
                {
                    throw new FaultException(string.Format("The topic-key '{0}' doesn't exist in the storage.", keyword),
                        FaultCode.CreateSenderFaultCode(WSEventing.InvalidMessage.SubCode, WSEventing.NamespaceUri));
                }
                else
                {
                    string key = (string)_storage.Key[keyword];
                    po = _storage.Val[key] as PersistedObject;
                }
            }
            else 
            {
                throw new FaultException(string.Format("Wrong ResourceDescriptor", keyword),
                    FaultCode.CreateSenderFaultCode(WSEventing.InvalidMessage.SubCode, WSEventing.NamespaceUri));
            }
            return po;
        }
        #endregion

        #region OnTimerCallback
        public void OnSubscriptionExpired(object sender, System.Timers.ElapsedEventArgs e)
        {
            try
            {
                _storage.rwl.AcquireWriterLock(TimeSpan.FromSeconds(10));
                PersistedObject po = sender as PersistedObject;
                if (po != null)
                {
                    po.Stop();
                    po.Close();
                    Delete<object>(po.ResourceDescriptor);

                    // raise event
                    if (this.ResourceDeleted != null)
                    {
                        ResourceDeletedEventArgs delevent = new ResourceDeletedEventArgs(po.ResourceDescriptor, po.Resource);
                        ResourceDeleted.Invoke(this, delevent);
                    }
                }
            }
            catch (Exception ex)
            {
                Trace.WriteLine(ex.Message);
            }
            finally
            {
                _storage.rwl.ReleaseWriterLock();
            }
        }
        #endregion
    }
}


