//*****************************************************************************
//    Description.....WS-Eventing for WCF
//                                
//    Author..........Roman Kiss, rkiss@pathcom.com
//    Copyright  2005 ATZ Consulting Inc. (see included license.rtf file)        
//                        
//    Date Created:    06/06/06
//
//    Date        Modified By     Description
//-----------------------------------------------------------------------------
//    06/06/06    Roman Kiss     Initial Revision
//    07/07/06    Roman Kiss     Version JuneCTP 2006
//    11/11/06    Roman Kiss     RTM
//*****************************************************************************
//  
#region References
using System;
using System.Diagnostics;
using System.Globalization;
using System.Threading;
using System.Configuration;
using System.Collections.Specialized;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Text;
using System.ServiceModel;
using System.Xml;
using System.IO;
using System.Transactions;
using System.Runtime.Serialization;
using System.Xml.Serialization;
using System.ServiceModel.Channels;
using System.ServiceModel.Description;
using System.ServiceModel.Dispatcher;
//
using RKiss.WSLib;
using RKiss.WSEventing.Adapters;
#endregion

namespace RKiss.WSEventing
{
    #region Service Contract(s)
    [ServiceContract]
    public interface IOneWay : IOutputChannel
    {
        [OperationContract(IsOneWay = true)]
        void OneWay(Message message);
    }

    [ServiceContract]
    public interface INotificationManager
    {
        [OperationContract(Action = "*", IsOneWay = true)]
        void FireSubscription(Message message);

        [OperationContract(Action = "CancelSubscription", IsOneWay = true)]
        void CancelSubscription(Uri code, string reason, string culturename);
    }
    #endregion

    /// <summary>
    /// Notification Manager Service
    /// </summary>
    //public class NotificationManagerService : NotificationManagerService<MemoryStorageAdapter> { }

    public class NotificationManagerService : NotificationManagerService<ConfiguredAddapter> { }

    [ServiceBehavior(IncludeExceptionDetailInFaults = true)]
    public class NotificationManagerService<T> : ServiceAdapterBase<T, IStorageAdapter>, INotificationManager where T : class
    {
        #region Constructor(s)
        public NotificationManagerService()
        {
            // features         
        }
        #endregion

        #region INotificationManager Members
        public void FireSubscription(Message message)
        {   
            // configuration
            string topic = Properties["topic"] as string;
            MessageProperties mp = OperationContext.Current.IncomingMessageProperties;
            Publishers publishers = mp.ContainsKey("EventingPublishers") ?
                mp["EventingPublishers"] as Publishers : null;
           
            using (TransactionScope ts = new TransactionScope(TransactionScopeOption.Required))
            {
                // Get subscriptions from Event Storage
                IList<Subscription> subscriptions = Adapter.Get<Subscription>(topic);

                // Create buffered message
                using (MessageBuffer buffer = message.CreateBufferedCopy(0xffffff))
                {
                    // walk through all subscriptions
                    foreach (Subscription subscription in subscriptions)
                    {
                        //Util.ShowMe(subscription);

                        // filter message
                        if (subscription.Filter == null || subscription.Filter.Match(buffer))
                        {
                            FireState state = new FireState(subscription, buffer.CreateMessage(), Properties);
                            Publisher publisher = null;

                            // delivery mode
                            if (publishers != null && publishers.TryGetValue(subscription.Delivery.DeliveryMode, out publisher))
                            {
                                // loosely coupled Publisher
                                Properties["Publisher"] = publisher;
                                PublisherBase driver = (PublisherBase)Activator.CreateInstance(publisher.Type, new object[] { Properties });
                                ThreadPool.QueueUserWorkItem(new WaitCallback(driver.DeliveryMessage), state);
                            }
                            else if (subscription.Delivery.DeliveryMode == null || subscription.Delivery.DeliveryMode == WSEventing.PushDeliveryMode)
                            {
                                /*Note:
                                 *  Unstable version ThreadPool.QueueUserWorkItem...multiple threads starting
                                 *  Stable version: FireSubscriptionWorker...only starting 1 at a time...
                                 */ 
                                // default built-in Push Publisher
                                //ThreadPool.QueueUserWorkItem(new WaitCallback(FireSubscriptionWorker), state);
                                FireSubscriptionWorker(state);
                            }
                        }
                    }
                }
            }
        }

        public void CancelSubscription(Uri code, string reason, string culturename)
        {
            string topic = Properties["topic"] as string;

            using (TransactionScope ts = new TransactionScope(TransactionScopeOption.Required))
            {
                IList<Subscription> subscriptions = Adapter.Delete<Subscription>(topic);
                foreach (Subscription subscription in subscriptions)
                {
                    if (subscription.EndTo == null) continue;
                    SubscriptionEnd subscriptionEnd = new SubscriptionEnd(subscription.Manager, Convert.ToSubscriptionEndCode(code),
                        reason, culturename == "" ? CultureInfo.CurrentCulture : new CultureInfo(culturename));
                    CancelState state = new CancelState(subscription.EndTo, subscriptionEnd, Properties);
                    ThreadPool.QueueUserWorkItem(new WaitCallback(CancelSubscriptionWorker), state);
                }
            }
        }
        #endregion

        #region Private Workers
        public void FireSubscriptionWorker(object state)
        {
            // state
            FireState firestate = state as FireState;

            try
            {
                // destination
                EndpointAddress epa = firestate.Subscription.Delivery.NotifyTo;

                // create configname name for configurable binding 
                string epname = (string)firestate.Properties["epname"] ?? string.Empty;
                string configname = string.Concat(epname, epa.Uri.Scheme);

                // action
                using (ChannelFactory<IOneWay> cf = new ChannelFactory<IOneWay>(configname, epa))
                {
                    IOutputChannel channel = (IOutputChannel)cf.CreateChannel();
                    channel.Send(firestate.Message);
                    channel.Close();
                }
            }
            catch (Exception ex)
            {
                try
                {
                    Adapter.Delete<Subscription>(firestate.Subscription.Manager.Identifier.Value);
                    SubscriptionEnd subscriptionEnd = new SubscriptionEnd(firestate.Subscription.Manager, SubscriptionEndCode.DeliveryFailure, ex.Message);
                    CancelState cancelstate = new CancelState(firestate.Subscription.EndTo, subscriptionEnd, firestate.Properties);
                    CancelSubscriptionWorker(cancelstate);

                    Trace.WriteLine(ex.Message);
                }
                catch (Exception ex2)
                {
                    Trace.WriteLine(ex2.Message);
                }
            }
        }

        public void CancelSubscriptionWorker(object state)
        {
            CancelState cancelstate = state as CancelState;

            try
            {
                // destination
                EndpointAddress epa = cancelstate.EndTo;

                // create configname name for configurable binding 
                string epname = (string)cancelstate.Properties["epname"] ?? string.Empty;
                string configname = string.Concat(epname, epa.Uri.Scheme);

                // action
                using (ChannelFactory<IOneWay> cf = new ChannelFactory<IOneWay>(configname, epa))
                {                   
                    Message message = Message.CreateMessage(MessageVersion.Soap12WSAddressing10, WSEventing.SubscriptionEndAction, cancelstate.SubscriptionEnd);
                    IOutputChannel channel = (IOutputChannel)cf.CreateChannel();
                    channel.Send(message);
                    channel.Close();
                }
            }
            catch (Exception ex)
            {
                // Nothing to do, just logging ex:
                Trace.WriteLine(ex.Message);
            }
        }
        #endregion   
    }
}
